//
//  CustomSlider.swift
//  LNPopupUIExample
//
//  Created by Vamsi Madduluri on 19/12/23.
//

import SwiftUI
import Combine

struct CustomSlider: View {
    @Binding var value: Double // Current playback time in seconds
    let totalDuration: Double // Total duration of the track in seconds
    let knobWidth: CGFloat = 20.0 // Width of the knob, increased for better touch target
    @State private var isEditing = false
    @Binding var buffering: Bool // Binding to the buffering state
    var onEditingChanged: (Bool) -> Void
    
    var body: some View {
        VStack {
            GeometryReader { geometry in
                ZStack(alignment: .leading) {
                    Rectangle()
                        .foregroundColor(Color.gray.opacity(0.3))
                        .frame(height: 4)
                    
                    Rectangle()
                        .foregroundColor(Color.white)
                        .frame(width: max(CGFloat(self.value / max(self.totalDuration, 1)) * geometry.size.width, 0),
                               height: 4)
                    
                    Circle()
                        .foregroundColor(Color.white)
                        .frame(width: self.knobWidth, height: self.knobWidth)
                        .offset(x: CGFloat(self.value / self.totalDuration) * geometry.size.width - self.knobWidth / 2)
                        .gesture(DragGesture(minimumDistance: 0)
                            .onChanged({ value in
                                self.isEditing = true
                                                let sliderPosition = value.location.x / geometry.size.width
                                                let sliderValue = Double(sliderPosition) * self.totalDuration
                                                self.value = min(max(sliderValue, 0), self.totalDuration) // Ensure value is within valid range
                                                self.onEditingChanged(true)
                            })
                                .onEnded({ _ in
                                    self.isEditing = false
                                    self.onEditingChanged(false)
                                })
                        ).animation(.interpolatingSpring(stiffness: 170, damping: 15), value: isEditing)
                }
                .frame(height: 20)
            }
            .frame(height: 20)
            
            HStack {
                // Current time label
                if buffering {
                    ProgressView()
                        .progressViewStyle(CircularProgressViewStyle())
                } else {
                    Text("\(formatTime(seconds: value))")
                        .foregroundColor(.white)
                        .font(.custom("Work Sans", size: 14))
                }
                Spacer()
                // Duration label
                Text("\(formatTime(seconds: totalDuration))")
                    .foregroundColor(.white)
                    .font(.custom("Work Sans", size: 14))
                
            }
        }
    }
    
    private func formatTime(seconds: Double) -> String {
        let time = Int(seconds)
        let minutes = time / 60
        let seconds = time % 60
        return String(format: "%02d:%02d", minutes, seconds)
    }
    
}

