//
//  MusicPlayerView.swift
//  LNPopupUIExample
//
//  Created by Vamsi Madduluri on 19/12/23.
//

import SwiftUI
import AVKit
import LNPopupUI
import SDWebImageSwiftUI

struct MusicPlayerView: View {
    @State private var albumArtwork: Image = Image(
        uiImage: .checkmark)
    @State private var transparancy = 0.0
    @ObservedObject var viewModel: MusicPlayerViewModel
    @State private var popupUIImage: UIImage? = nil
    
    var body: some View {
        ZStack {
            viewModel.gradientBackground
            VStack(spacing: 80) {
                // Close down chevron
                HStack {
                    Button(action: {
                        viewModel.pause()
                        viewModel.isPopupOpen = false
                    }) {
                        Image(systemName: "chevron.down")
                            .font(.system(size: 24))
                            .foregroundColor(.white)
                    }
                    Spacer()
                    Spacer()
                }.padding(.top, 70).padding(.horizontal, 24)
                
                VStack {
                    if let imgUrl = viewModel.imgUrl {
                        AnimatedImage(url: URL(string: imgUrl))
                            .onSuccess { image, data, cacheType in
                                // This is where you receive the UIImage
                                self.viewModel.updateGradientBackground(from: image)
                            }
                            .resizable()
                            .aspectRatio(contentMode: .fill)
                            .frame(width: 340, height: 340)
                            .cornerRadius(10).padding(.bottom)
                            .animation(.interpolatingSpring(stiffness: 170, damping: 15), value: viewModel.isPlaying)
                        
                    }
                    
                    if let sermonsData = viewModel.sermonsData {
                        VStack(spacing: 6) {
                            Text(sermonsData.title)
                                .font(.custom("Work Sans", size: 24))
                                .fontWeight(.bold)
                            
                            Text(sermonsData.preacherName ?? "")
                                .font(.custom("Work Sans", size: 18))
                                .fontWeight(.regular)
                                .foregroundColor(.secondary)
                        }
                    }
                   
                    CustomSlider(
                        value: Binding(
                            get: { viewModel.progress * viewModel.totalDuration },
                            set: { newValue in viewModel.progress = newValue / viewModel.totalDuration }
                        ),
                        totalDuration: viewModel.totalDuration,
                        buffering: $viewModel.buffering,
                        onEditingChanged: { isEditing in
                            viewModel.seekToTime(isEditing: isEditing)
                        }
                    )
                    .padding(.top).padding(.horizontal, 38)
                    
                    
                    // Playback controls
                    playbackControls
                        .animation(.easeInOut, value: viewModel.isPlaying)
                }
                Spacer()
            }.onAppear {
                loadUIImageFromURL(artwork: viewModel.sermonsData?.artwork)
            }.onDisappear {
                viewModel.pause()
            }
            .onReceive(viewModel.$sermonsData) { currentPlaybackData in
                loadUIImageFromURL(artwork: currentPlaybackData?.artwork)
            }.popupBarItems({
                Button(action: {
                    viewModel.togglePlayPause()
                }) {
                    Image(systemName: viewModel.isPlaying ? "pause.fill" : "play.fill")
                }
                
            }).popupTitle(viewModel.sermonsData?.title ?? "Not Playing")
                .popupImage(popupUIImage ?? UIImage(systemName: "photo")!)
        }.edgesIgnoringSafeArea(.all)
    }
    
    private func loadUIImageFromURL(artwork: String?) {
        
        guard let artwork = artwork,
              let artworkURL = URL(string: "https://dev1-directus.ihdxht.easypanel.host/assets/\(artwork)") else {
            self.popupUIImage = UIImage(systemName: "photo")
            return
        }
        
        URLSession.shared.dataTask(with: artworkURL) { data, _, _ in
            if let data = data, let image = UIImage(data: data) {
                DispatchQueue.main.async {
                    self.popupUIImage = image
                }
            } else {
                DispatchQueue.main.async {
                    self.popupUIImage = UIImage(systemName: "photo")
                }
            }
        }.resume()
    }
    
    var playbackControls: some View {
        HStack(spacing: 30) {
            Button(action: {
                viewModel.skipBackward()
            }) {
                Image(systemName: "gobackward.30")
                
            }
            
            Button(action: {
                viewModel.togglePlayPause()
                transparancy = 0.6
                withAnimation(.easeInOut(duration: 0.4)) {
                    DispatchQueue.main.asyncAfter(deadline: .now() + 0.4) {
                        transparancy = 0.0
                    }
                }
            }) {
                ZStack {
                    
                    Circle()
                        .frame(width: 90, height: 90)
                        .opacity(transparancy)
                    
                    Image(systemName: "play.fill")
                        .font(.system(size: 54))
                        .scaleEffect(viewModel.isPlaying ? 0 : 1)
                        .animation(.interpolatingSpring(stiffness: 170, damping: 15), value: viewModel.isPlaying)
                    
                    
                    Image(systemName: "pause.fill")
                        .font(.system(size: 54))
                        .scaleEffect(viewModel.isPlaying ? 1 : 0)
                        .animation(.interpolatingSpring(stiffness: 170, damping: 15), value: viewModel.isPlaying)
                    
                    
                }
            }.disabled(viewModel.buffering)
            
            Button(action: {
                // Next track action
                viewModel.skipForward()
            }) {
                Image(systemName: "goforward.30")
                
            }
        }
        .font(.title)
        .padding()
    }
}

extension UIImage {
    func averageColor() -> UIColor? {
        guard let inputImage = CIImage(image: self) else { return nil }
        let extentVector = CIVector(x: inputImage.extent.origin.x, y: inputImage.extent.origin.y, z: inputImage.extent.size.width, w: inputImage.extent.size.height)
        
        guard let filter = CIFilter(name: "CIAreaAverage", parameters: [kCIInputImageKey: inputImage, kCIInputExtentKey: extentVector]) else { return nil }
        guard let outputImage = filter.outputImage else { return nil }
        
        var bitmap = [UInt8](repeating: 0, count: 4)
        let context = CIContext(options: [.workingColorSpace: kCFNull!])
        context.render(outputImage, toBitmap: &bitmap, rowBytes: 4, bounds: CGRect(x: 0, y: 0, width: 1, height: 1), format: .RGBA8, colorSpace: nil)
        
        return UIColor(red: CGFloat(bitmap[0]) / 255, green: CGFloat(bitmap[1]) / 255, blue: CGFloat(bitmap[2]) / 255, alpha: CGFloat(bitmap[3]) / 255)
    }
}

struct AudioSermonData: Codable {
    let id: String
    let status: String
    let sort: String?
    let userCreated: String?
    let dateCreated: String?
    let userUpdated: String?
    let dateUpdated: String?
    let type: String
    let streamUrl: String?
    let duration: Int
    let title: String
    let preacherId: String
    var preacherName: String?
    let description: String
    let sermonSeries: Bool?
    let sermonOfTheDay: Bool
    let artwork: String
}

struct AudioSermonListResponse: Codable {
    let data: [AudioSermonData]
}

#Preview {
    MusicPlayerView(viewModel: MusicPlayerViewModel())
}
