//
//  MusicPlayerViewModel.swift
//  LNPopupUIExample
//
//  Created by Vamsi Madduluri on 19/12/23.
//

import SwiftUI
import Combine
import AVKit
import MediaPlayer

class MusicPlayerViewModel: NSObject, ObservableObject {
    @Published var isPopupBarPresented: Bool = false
    @Published var isPopupOpen: Bool = false
    @Published var isPlaying = false
    @Published var progress = 0.0
    @Published var totalDuration = 0.0
    @Published var gradientBackground: LinearGradient = LinearGradient(gradient: Gradient(colors: [.black]), startPoint: .top, endPoint: .bottom)
    @Published var buffering = true
    @Published var isSeeking = false
    var player: AVPlayer?
    var timeObserverToken: Any?
    var playerItem: AVPlayerItem?
    @Published var imgUrl: String?
    @Published var audioUrl: String?
    @Published var sermonsData: AudioSermonData?
    // Call this method at appropriate times to update the buffering state
    func updateBufferingState(isBuffering: Bool) {
        DispatchQueue.main.async {
            withAnimation {
                self.buffering = isBuffering
            }
        }
    }
    @Published var popupUIImage: UIImage?
    
    override init() {
        super.init()
        setupAudioSession()
        setupPlayerNotifications()
        configureRemoteCommandCenter()
    }
    
    private func configureRemoteCommandCenter() {
        let commandCenter = MPRemoteCommandCenter.shared()
        
        // Play command
        commandCenter.playCommand.addTarget { [weak self] event in
            self?.play()
            return .success
        }
        
        // Pause command
        commandCenter.pauseCommand.addTarget { [weak self] event in
            self?.pause()
            return .success
        }
        
        // Skip forward command (e.g., 30 seconds)
        commandCenter.skipForwardCommand.preferredIntervals = [30] // Set preferred skip interval
        commandCenter.skipForwardCommand.addTarget { [weak self] event in
            self?.skipForward()
            return .success
        }
        
        // Skip backward command (e.g., 30 seconds)
        commandCenter.skipBackwardCommand.preferredIntervals = [30] // Set preferred skip interval
        commandCenter.skipBackwardCommand.addTarget { [weak self] event in
            self?.skipBackward()
            return .success
        }
    }
    
    func play() {
        player?.play()
        isPlaying = true
        updateNowPlayingInfo() // Update the Now Playing Info to reflect the change
    }
    
    func pause() {
        player?.pause()
        isPlaying = false
        updateNowPlayingInfo() // Update the Now Playing Info to reflect the change
    }
    
    private func setupAudioSession() {
        do {
            try AVAudioSession.sharedInstance().setCategory(.playback)
            try AVAudioSession.sharedInstance().setActive(true)
        } catch {
            print("Failed to set up audio session: \(error)")
        }
    }
    
    private func setupPlayerNotifications() {
        NotificationCenter.default.addObserver(self,
                                               selector: #selector(handleInterruption),
                                               name: AVAudioSession.interruptionNotification,
                                               object: AVAudioSession.sharedInstance())
        
        NotificationCenter.default.addObserver(self,
                                               selector: #selector(handleRouteChange),
                                               name: AVAudioSession.routeChangeNotification,
                                               object: AVAudioSession.sharedInstance())
    }
    
    @objc private func handleInterruption(notification: Notification) {
        guard let userInfo = notification.userInfo,
              let typeValue = userInfo[AVAudioSessionInterruptionTypeKey] as? UInt,
              let type = AVAudioSession.InterruptionType(rawValue: typeValue) else {
            return
        }
        
        if type == .began {
            // Interruption began, pause the player
            player?.pause()
            isPlaying = false
        } else if type == .ended {
            guard let optionsValue = userInfo[AVAudioSessionInterruptionOptionKey] as? UInt else {
                return
            }
            let options = AVAudioSession.InterruptionOptions(rawValue: optionsValue)
            if options.contains(.shouldResume) {
                // Interruption ended. Resume playback, if appropriate
                player?.play()
                isPlaying = true
            }
            DispatchQueue.main.async {
                self.updateNowPlayingInfo()
            }
        }
    }
    
    // Skip forward by 30 seconds
    func skipForward() {
        guard let currentItem = player?.currentItem else { return }
        let playerCurrentTime = CMTimeGetSeconds(player?.currentTime() ?? CMTime.zero)
        let duration = CMTimeGetSeconds(currentItem.duration)
        let newTime = playerCurrentTime + 30.0
        
        if newTime < duration {
            let time: CMTime = CMTimeMake(value: Int64(newTime * 1000 as Float64), timescale: 1000)
            player?.seek(to: time, toleranceBefore: .zero, toleranceAfter: .zero)
        }
    }
    
    // Skip backward by 30 seconds
    func skipBackward() {
        let playerCurrentTime = CMTimeGetSeconds(player?.currentTime() ?? CMTime.zero)
        var newTime = playerCurrentTime - 30.0
        
        if newTime < 0 {
            newTime = 0
        }
        
        let time: CMTime = CMTimeMake(value: Int64(newTime * 1000 as Float64), timescale: 1000)
        player?.seek(to: time, toleranceBefore: .zero, toleranceAfter: .zero)
    }
    
    
    func fetchAndSetImage(from urlString: String) {
        guard let url = URL(string: urlString) else { return }
        
        URLSession.shared.dataTask(with: url) { [weak self] data, _, _ in
            guard let data = data, let image = UIImage(data: data) else { return }
            DispatchQueue.main.async {
                self?.popupUIImage = image
                self?.updateNowPlayingInfo()
            }
        }.resume()
    }
    
    @objc private func handleRouteChange(notification: Notification) {
        guard let userInfo = notification.userInfo,
              let reasonValue = userInfo[AVAudioSessionRouteChangeReasonKey] as? UInt,
              let reason = AVAudioSession.RouteChangeReason(rawValue:reasonValue) else {
            return
        }
        
        switch reason {
        case .newDeviceAvailable: // e.g., headphones plugged in
            // Handle new device available
            break
        case .oldDeviceUnavailable: // e.g., headphones unplugged
            if let previousRoute = userInfo[AVAudioSessionRouteChangePreviousRouteKey] as? AVAudioSessionRouteDescription,
               previousRoute.outputs.contains(where: { $0.portType == .headphones }) {
                // Headphones were unplugged. Pause playback
                player?.pause()
                isPlaying = false
            }
        default: break
        }
    }
    
    func setupPlayer(with url: URL) {
        // Check if the current URL is different from the new URL
        if let currentUrl = playerItem?.asset as? AVURLAsset, currentUrl.url == url {
            // URL is the same, do not change the player item
            return
        }
        
        // URL is different, create a new player item
        self.playerItem = AVPlayerItem(url: url)
        self.player = AVPlayer(playerItem: playerItem)
        addObserverToPlayerItem(playerItem!)
        addPeriodicTimeObserver()
        if let imgUrl = imgUrl {
            fetchAndSetImage(from: imgUrl)
        }
        // Check if the popup bar is not presented, and present it if necessary
        if !isPopupBarPresented {
            isPopupBarPresented = true
            isPopupOpen = true
            if let imgUrl = imgUrl {
                fetchAndSetImage(from: imgUrl)
            }
        } else if let imgUrl = imgUrl, popupUIImage == nil {
            // Only fetch the image if it hasn't been set yet
            fetchAndSetImage(from: imgUrl)
        }
        updateNowPlayingInfo() // Always update now playing info after setting up
    }
    
    
    func updateNowPlayingInfo() {
        guard let currentItem = player?.currentItem,
                  currentItem.status == .readyToPlay else { return }
        var nowPlayingInfo = [String: Any]()
        guard let currentPlaybackData = sermonsData else { return }
        nowPlayingInfo[MPMediaItemPropertyTitle] = currentPlaybackData.title
        nowPlayingInfo[MPMediaItemPropertyArtist] = currentPlaybackData.preacherName
        nowPlayingInfo[MPNowPlayingInfoPropertyPlaybackRate] = isPlaying ? 1.0 : 0.0
        
        // If there's an existing image, we set it right away.
        if let popupImage = self.popupUIImage {
            let artwork = MPMediaItemArtwork(boundsSize: popupImage.size) { _ in popupImage }
            nowPlayingInfo[MPMediaItemPropertyArtwork] = artwork
        }
        
        let urlString = "https://dev1-directus.ihdxht.easypanel.host/\(currentPlaybackData.artwork)"
        nowPlayingInfo[MPNowPlayingInfoPropertyElapsedPlaybackTime] = player?.currentTime().seconds
        nowPlayingInfo[MPMediaItemPropertyPlaybackDuration] = player?.currentItem?.duration.seconds
        
        MPNowPlayingInfoCenter.default().nowPlayingInfo = nowPlayingInfo
    }
    
    func playbackStateChanged() {
        updateNowPlayingInfo()
    }
    
    func addObserverToPlayerItem(_ playerItem: AVPlayerItem) {
        playerItem.addObserver(self, forKeyPath: #keyPath(AVPlayerItem.status), options: [.old, .new], context: nil)
        playerItem.addObserver(self, forKeyPath: #keyPath(AVPlayerItem.loadedTimeRanges), options: [.old, .new], context: nil)
        playerItem.addObserver(self, forKeyPath: #keyPath(AVPlayerItem.isPlaybackBufferEmpty), options: [.old, .new], context: nil)
        playerItem.addObserver(self, forKeyPath: #keyPath(AVPlayerItem.isPlaybackLikelyToKeepUp), options: [.old, .new], context: nil)
    }
    
    override func observeValue(forKeyPath keyPath: String?, of object: Any?, change: [NSKeyValueChangeKey : Any]?, context: UnsafeMutableRawPointer?) {
        if object is AVPlayerItem {
            switch keyPath {
            case #keyPath(AVPlayerItem.status):
                if player?.status == .readyToPlay {
                    if let duration = player?.currentItem?.duration.seconds, !duration.isNaN {
                        DispatchQueue.main.async {
                            self.totalDuration = duration
                        }
                    }
                }
            case #keyPath(AVPlayerItem.loadedTimeRanges):
                let loadedTimeRanges = player?.currentItem?.loadedTimeRanges
                let isBuffering = loadedTimeRanges?.first?.timeRangeValue.end.seconds ?? 0 < player?.currentTime().seconds ?? 0
                DispatchQueue.main.async {
                    self.buffering = isBuffering
                }
            case #keyPath(AVPlayerItem.isPlaybackBufferEmpty):
                DispatchQueue.main.async {
                    self.buffering = true
                }
            case #keyPath(AVPlayerItem.isPlaybackLikelyToKeepUp):
                DispatchQueue.main.async {
                    self.buffering = false
                }
            default:
                break
            }
        }
    }
    func seekToTime(isEditing: Bool) {
        guard !isEditing, !buffering else { return }
        let targetTime = CMTime(seconds: min(max(progress * totalDuration, 0), totalDuration), preferredTimescale: 600)
        player?.seek(to: targetTime, toleranceBefore: .zero, toleranceAfter: .zero) { [weak self] _ in
            // Optionally resume playback here if it was paused during editing
            if self?.isPlaying == true {
                self?.player?.play()
            }
        }
    }
    
    // Modify your periodic time observer to update less frequently or only when not seeking
    func addPeriodicTimeObserver() {
        let interval = CMTime(seconds: 1, preferredTimescale: CMTimeScale(NSEC_PER_SEC))
        timeObserverToken = player?.addPeriodicTimeObserver(forInterval: interval, queue: .main) { [weak self] time in
            guard let self = self, !self.isSeeking else { return }
            let newProgress = time.seconds / (self.player?.currentItem?.duration.seconds ?? 1)
            // Update only if the difference is significant
            if abs(newProgress - self.progress) > 0.01 {
                DispatchQueue.main.async {
                    self.progress = newProgress
                }
            }
        }

    }
    
    
    deinit {
        if let token = timeObserverToken {
            player?.removeTimeObserver(token)
        }
        playerItem?.removeObserver(self, forKeyPath: #keyPath(AVPlayerItem.status))
        playerItem?.removeObserver(self, forKeyPath: #keyPath(AVPlayerItem.loadedTimeRanges))
        playerItem?.removeObserver(self, forKeyPath: #keyPath(AVPlayerItem.isPlaybackBufferEmpty))
        playerItem?.removeObserver(self, forKeyPath: #keyPath(AVPlayerItem.isPlaybackLikelyToKeepUp))
    }
    
    func startPlayback() {
        player?.play()
        isPlaying = true
        addPeriodicTimeObserver()
        updateTotalDuration()
        updateNowPlayingInfo()
    }
    
    // Add logic to update the `buffering` state based on player status
    // You might need to observe the player's status and loading times to properly update this state
    // This is a simplified example
    private func checkBuffering() {
        guard let currentItem = player?.currentItem else { return }
        
        // This is a simplified check, replace with actual buffering logic
        let isBuffering = currentItem.isPlaybackLikelyToKeepUp == false
        updateBufferingState(isBuffering: isBuffering)
    }
    
    
    // Update this method to handle playback state changes
    func togglePlayPause() {
        // Check if buffering before toggling playback state
        guard !buffering else { return }
        isPlaying.toggle()
        if isPlaying {
            player?.play()
        } else {
            player?.pause()
        }
        updateNowPlayingInfo()
    }
    
    
    private func updateTotalDuration() {
        if let duration = player?.currentItem?.duration.seconds, !duration.isNaN {
            totalDuration = duration
        }
    }
    
    // This function is called to update the gradient background based on an image.
    func updateGradientBackground(from image: UIImage) {
        let gradientColors = getGradientColors(from: image)
        DispatchQueue.main.async {
            withAnimation {
                self.gradientBackground = LinearGradient(
                    gradient: Gradient(colors: gradientColors.map { Color($0) }),
                    startPoint: .top,
                    endPoint: .bottom
                )
            }
        }
    }
    
    // This function extracts colors to create a gradient background.
    private func getGradientColors(from image: UIImage) -> [UIColor] {
        // Resize image to reduce the number of pixels to analyze
        let size = CGSize(width: 50, height: 50) // Reducing the size to speed up processing
        UIGraphicsBeginImageContext(size)
        image.draw(in: CGRect(origin: .zero, size: size))
        guard let resizedImage = UIGraphicsGetImageFromCurrentImageContext() else { return [.black] }
        UIGraphicsEndImageContext()
        
        guard let cgImage = resizedImage.cgImage else { return [.black] }
        let width = cgImage.width
        let height = cgImage.height
        let colorSpace = CGColorSpaceCreateDeviceRGB()
        var rawData = [UInt8](repeating: 0, count: width * height * 4)
        let bytesPerPixel = 4
        let bytesPerRow = bytesPerPixel * width
        let bitsPerComponent = 8
        
        let bitmapInfo = CGImageAlphaInfo.premultipliedLast.rawValue | CGBitmapInfo.byteOrder32Big.rawValue
        guard let context = CGContext(data: &rawData, width: width, height: height, bitsPerComponent: bitsPerComponent, bytesPerRow: bytesPerRow, space: colorSpace, bitmapInfo: bitmapInfo) else { return [.black] }
        
        context.draw(cgImage, in: CGRect(x: 0, y: 0, width: width, height: height))
        
        // Analyze pixels to find colors
        var colorMap: [UIColor: Int] = [:]
        for y in 0..<height {
            for x in 0..<width {
                let byteIndex = (bytesPerRow * y) + x * bytesPerPixel
                let red   = CGFloat(rawData[byteIndex]) / 255.0
                let green = CGFloat(rawData[byteIndex + 1]) / 255.0
                let blue  = CGFloat(rawData[byteIndex + 2]) / 255.0
                let alpha = CGFloat(rawData[byteIndex + 3]) / 255.0
                let color = UIColor(red: red, green: green, blue: blue, alpha: alpha)
                // Exclude white and near-white colors, focus on colors with lower brightness
                if color.isDarkColor {
                    colorMap[color, default: 0] += 1
                }
            }
        }
        
        // Sort the colors by frequency
        let sortedColors = colorMap.sorted { $0.value > $1.value }.map { $0.key }
        let selectedColors = sortedColors.prefix(3) // Take the top 3 colors
        return Array(selectedColors)
    }
    
}


extension UIColor {
    var brightness: CGFloat {
        var brightness: CGFloat = 0
        self.getHue(nil, saturation: nil, brightness: &brightness, alpha: nil)
        return brightness
    }
}

extension UIColor {
    // A computed property to get brightness value based on the HSB color model
    var perceivedBrightness: CGFloat {
        var brightness: CGFloat = 0
        self.getHue(nil, saturation: nil, brightness: &brightness, alpha: nil)
        return brightness
    }
}


// MARK: - UIColor Extension
extension UIColor {
    var isDarkColor: Bool {
        var white: CGFloat = 0
        self.getWhite(&white, alpha: nil)
        return white < 0.5 // Adjust threshold as needed
    }
}
