//
//  NetworkFetcher.swift
//  LNPopupUIExample
//
//  Created by Vamsi Madduluri on 20/12/23.
//

import Foundation
import RealmSwift

protocol IdentifiableObject: Object, Codable, Equatable {
    var id: String { get }
}

protocol Mappable {
    associatedtype RealmType: Object
    func mapToRealmModel() -> RealmType
}

struct GenericResponse<T: Codable>: Codable {
    let data: T
}


class NetworkFetcher {
    static let shared = NetworkFetcher()
    private init() {}
    @MainActor
    func fetchData<T: Codable>(
        using networkingManager: NetworkingManagerImpl,
        endpoint: Endpoint,
        resetAction: @escaping () -> Void,
        updateState: @escaping (T) -> Void,
        handleError: @escaping (Error) -> Void
    ) async {
        resetAction()
        do {
            let response: T = try await networkingManager.request(session: .shared, endpoint, type: T.self)
            updateState(response)
        } catch {
            handleError(error)
        }
    }
    
    
    @MainActor
    func fetchData<T: Object & Codable, MappableType: Codable>(
        using networkingManager: NetworkingManagerImpl,
        endpoint: Endpoint,
        resetAction: @escaping () -> Void,
        updateState: @escaping ([T], Bool) -> Void,
        handleError: @escaping (Error) -> Void,
        mapFunction: ((MappableType) -> T)? = nil
    ) async {
        resetAction()

        do {
            let wrappedResponse: GenericResponse<[MappableType]> = try await networkingManager.request(session: .shared, endpoint, type: GenericResponse<[MappableType]>.self)
            let response = wrappedResponse.data

            let finalData: [T] = mapFunction != nil ? response.map { mapFunction!($0) } : response as! [T]

            let realmObjects = RealmManager.shared.getObjects(ofType: T.self)
            let realmDataArray = Array(realmObjects)

            // Check for objects to delete
            let objectsToDelete = realmDataArray.filter { !finalData.contains($0) }
            RealmManager.shared.delete(objectsToDelete)

            if realmDataArray != finalData {
                finalData.forEach { RealmManager.shared.add($0, update: .modified) }
                updateState(finalData, false)
            } else {
                updateState(realmDataArray, true)
            }
        } catch {
            let realmDataArray = Array(RealmManager.shared.getObjects(ofType: T.self))
            updateState(realmDataArray, true)
            handleError(error)
        }
    }
    
    
    @MainActor
    func fetchDataDirectly<T: Object & Codable>(
        using networkingManager: NetworkingManagerImpl,
        endpoint: Endpoint,
        resetAction: @escaping () -> Void,
        updateState: @escaping ([T], Bool) -> Void,
        handleError: @escaping (Error) -> Void
    ) async {
        resetAction()

        do {
            let wrappedResponse: GenericResponse<[T]> = try await networkingManager.request(session: .shared, endpoint, type: GenericResponse<[T]>.self)
            let response = wrappedResponse.data

            let realmObjects = RealmManager.shared.getObjects(ofType: T.self)
            let realmDataArray = Array(realmObjects)

            // Check for objects to delete
            let objectsToDelete = realmDataArray.filter { !response.contains($0) }
            RealmManager.shared.delete(objectsToDelete)

            // Check for objects to add or update
            let objectsToAddOrUpdate = response.filter { !realmDataArray.contains($0) }
            objectsToAddOrUpdate.forEach { RealmManager.shared.add($0, update: .modified) }

            if objectsToAddOrUpdate.isEmpty {
                updateState(realmDataArray, true)
            } else {
                updateState(response, false)
            }
        } catch {
            let realmDataArray = Array(RealmManager.shared.getObjects(ofType: T.self))
            updateState(realmDataArray, true)
            handleError(error)
        }
    }
    
    // Method for uploading files
    @MainActor
    func uploadFile<T: Codable>(
        using networkingManager: NetworkingManagerImpl,
        endpoint: Endpoint,
        fileURL: URL,
        parameters: [String: String], // Adjusted to [String: String]
        resetAction: @escaping () -> Void,
        updateState: @escaping (T) -> Void,
        handleError: @escaping (Error) -> Void
    ) async {
        resetAction()
        do {
            let response: T = try await networkingManager.uploadFile(
                session: .shared,
                endpoint: endpoint,
                file: fileURL,
                parameters: parameters, // Passed as [String: String]
                type: T.self
            )
            updateState(response)
        } catch {
            handleError(error)
        }
    }
}

struct EmptyResponse: Codable {}
