//
//  NetworkingManagerImpl.swift
//  LNPopupUIExample
//
//  Created by Vamsi Madduluri on 19/12/23.
//

import Foundation

protocol NetworkingManagerImpl {
    func request<T: Codable>(session: URLSession, _ endpoint: Endpoint, type: T.Type) async throws -> T
    func request(session: URLSession, _ endpoint: Endpoint) async throws
    func uploadFiles(session: URLSession, endpoint: Endpoint, files: [URL], parameters: [String: String]) async throws
    func uploadFile<T: Codable>(session: URLSession, endpoint: Endpoint, file: URL, parameters: [String: String], type: T.Type) async throws -> T
}


@MainActor
final class NetworkingManager: NetworkingManagerImpl {
    static let shared = NetworkingManager()
    private var webSocketTask: URLSessionWebSocketTask?
    
    private init() {
        print("NetworkingManager initialized")
    }
    
    
    func request<T: Codable>(session: URLSession = .shared, _ endpoint: Endpoint, type: T.Type) async throws -> T {
        guard let url = endpoint.url else {
            print("Invalid URL for endpoint: \(endpoint)")
            throw NetworkingError.invalidUrl
        }
        
        var request = buildRequest(from: url, methodType: endpoint.methodType)
        print("Constructed request for URL: \(url.absoluteString)")
        
        let (data, response) = try await session.data(for: request)
        guard let httpResponse = response as? HTTPURLResponse else {
            throw NetworkingError.invalidResponse
        }
        
        print(response)
        
        if (200...299).contains(httpResponse.statusCode) {
            let decoder = JSONDecoder()
            decoder.dateDecodingStrategy = .formatted(dateFormatter)
            decoder.keyDecodingStrategy = .convertFromSnakeCase
            return try decoder.decode(T.self, from: data)
        } else {
            if (try? JSONDecoder().decode(ServerErrorResponse.self, from: data)) != nil {
                throw NetworkingError.invalidStatusCode(statusCode: httpResponse.statusCode)
            } else {
                throw NetworkingError.invalidStatusCode(statusCode: httpResponse.statusCode)
            }
        }
    }
    
    func request(session: URLSession = .shared, _ endpoint: Endpoint) async throws {
        
        guard let url = endpoint.url else {
            print("NetworkingError: Invalid URL")
            throw NetworkingError.invalidUrl
        }
        
        var request = buildRequest(from: url, methodType: endpoint.methodType)
        print("Constructed request for URL: \(url.absoluteString), Method: \(endpoint.methodType)")
        
        
        do {
            let (data, response) = try await session.data(for: request)
            
            guard let httpResponse = response as? HTTPURLResponse else {
                print("NetworkingError: Invalid response")
                throw NetworkingError.invalidResponse
            }
            
            print("Response received: URL: \(url.absoluteString), Status Code: \(httpResponse.statusCode)")
            
            if !(200...299).contains(httpResponse.statusCode) {
                let responseBody = String(data: data, encoding: .utf8) ?? "No response body"
                print("NetworkingError: Invalid status code: \(httpResponse.statusCode), Response Body: \(responseBody)")
                throw NetworkingError.invalidStatusCode(statusCode: httpResponse.statusCode)
            }
            
        } catch {
            print("NetworkingError: Exception during request: \(error)")
            throw error
        }
    }
    
    func uploadFiles(session: URLSession = .shared, endpoint: Endpoint, files: [URL], parameters: [String: String]) async throws {
        guard let url = endpoint.url else {
            throw NetworkingError.invalidUrl
        }
        
        let boundary = "Boundary-\(UUID().uuidString)"
        var request = buildRequest(from: url, methodType: endpoint.methodType, boundary: boundary)
        
       
        
        request.httpBody = createBody(boundary: boundary, parameters: parameters, files: files)
        
        let (_, response) = try await session.data(for: request)
        
        guard let httpResponse = response as? HTTPURLResponse, (200...299).contains(httpResponse.statusCode) else {
            throw NetworkingError.invalidStatusCode(statusCode: (response as? HTTPURLResponse)?.statusCode ?? 0)
        }
    }
    
    func uploadFile<T: Codable>(session: URLSession = .shared, endpoint: Endpoint, file: URL, parameters: [String: String], type: T.Type) async throws -> T {
        guard let url = endpoint.url else {
            print("NetworkingError: Invalid URL")
            throw NetworkingError.invalidUrl
        }
        
        let boundary = "Boundary-\(UUID().uuidString)"
        var request = buildRequest(from: url, methodType: endpoint.methodType, boundary: boundary)

        request.httpBody = createBody(boundary: boundary, parameters: parameters, files: [file])
        
        // Log the request details
        print("Request URL: \(url)")
        print("Method: \(endpoint.methodType)")
        print("Headers: \(request.allHTTPHeaderFields ?? [:])")
        if let httpBody = request.httpBody, let bodyString = String(data: httpBody, encoding: .utf8) {
            print("HTTP Body: \(bodyString)")
        }
        
        do {
            let (data, response) = try await session.data(for: request)
            
            guard let httpResponse = response as? HTTPURLResponse, (200...299).contains(httpResponse.statusCode) else {
                print("Error: Invalid status code: \((response as? HTTPURLResponse)?.statusCode ?? 0)")
                if let responseData = String(data: data, encoding: .utf8) {
                    print("Response: \(responseData)")
                }
                throw NetworkingError.invalidStatusCode(statusCode: (response as? HTTPURLResponse)?.statusCode ?? 0)
            }
            
            let decoder = JSONDecoder()
            decoder.dateDecodingStrategy = .formatted(dateFormatter)
            decoder.keyDecodingStrategy = .convertFromSnakeCase
            return try decoder.decode(T.self, from: data)
        } catch {
            print("Exception during file upload: \(error)")
            throw error
        }
    }
}

extension NetworkingManager {
    enum NetworkingError: LocalizedError {
        case invalidUrl
        case custom(error: Error)
        case invalidStatusCode(statusCode: Int)
        case invalidResponse
        case serverError(code: Int, error: ServerErrorResponse)
        case invalidData
        case failedToDecode(error: Error)
    }
}

extension NetworkingManager.NetworkingError: Equatable {
    static func == (lhs: NetworkingManager.NetworkingError, rhs: NetworkingManager.NetworkingError) -> Bool {
        switch (lhs, rhs) {
        case (.invalidUrl, .invalidUrl), (.invalidResponse, .invalidResponse), (.invalidData, .invalidData):
            return true
        case (.custom(let lhsError), .custom(let rhsError)):
            return lhsError.localizedDescription == rhsError.localizedDescription
        case (.invalidStatusCode(let lhsCode), .invalidStatusCode(let rhsCode)):
            return lhsCode == rhsCode
        case (.serverError(let lhsCode, _), .serverError(let rhsCode, _)):
            return lhsCode == rhsCode
        case (.failedToDecode(let lhsError), .failedToDecode(let rhsError)):
            return lhsError.localizedDescription == rhsError.localizedDescription
        default:
            return false
        }
    }
}

extension NetworkingManager.NetworkingError {
    var userFriendlyDescription: String? {
        switch self {
        case .invalidUrl:
            return "We couldn't find the requested resource. Please try again."
        case .invalidStatusCode(let code):
            return "There was an unexpected error while processing your request. Please try again later. (Code: \(code))"
        case .invalidResponse:
            return "We received an unexpected response. Please try again later."
        case .serverError(_, let error):
            return "The server encountered an error. Please try again later. (Error: \(error.errors.first?.message ?? "Unknown error"))"
        case .invalidData:
            return "We received corrupted data. Please try again later."
        case .failedToDecode(let error):
            return "We couldn't understand the received information. Please try again later. (Error: \(error.localizedDescription))"
        case .custom(let error):
            return "Something unexpected happened. Please try again later. (Error: \(error.localizedDescription))"
        }
    }
}


private extension NetworkingManager {
    var dateFormatter: DateFormatter {
        let formatter = DateFormatter()
        formatter.dateFormat = "yyyy-MM-dd'T'HH:mm:ss"
        formatter.locale = Locale(identifier: "en_US_POSIX")
        return formatter
    }
}

private extension NetworkingManager {
    func buildRequest(from url: URL, methodType: Endpoint.MethodType, boundary: String? = nil) -> URLRequest {
        var request = URLRequest(url: url)
        
        switch methodType {
        case .GET:
            request.httpMethod = "GET"
            
        case .POST(let data), .PATCH(let data):
            request.httpMethod = methodType.methodName
            request.setValue("application/json", forHTTPHeaderField: "Content-Type")
            request.httpBody = data
            
        case .UPLOAD:
            request.httpMethod = "POST"
            if let boundary = boundary {
                request.setValue("multipart/form-data; boundary=\(boundary)", forHTTPHeaderField: "Content-Type")
            }
            
        case .DELETE:
            request.httpMethod = "DELETE"
            
        }
        
        return request
    }
}



private extension NetworkingManager {
    func createBody(boundary: String, parameters: [String: String], files: [URL]) -> Data {
        var body = Data()
        
        for (key, value) in parameters {
            body.append("--\(boundary)\r\n")
            body.append("Content-Disposition: form-data; name=\"\(key)\"\r\n\r\n")
            body.append("\(value)\r\n")
        }
        
        for fileUrl in files {
            guard let fileData = try? Data(contentsOf: fileUrl),
                  let mimeType = mimeTypeFor(fileUrl.pathExtension) else {
                continue
            }
            
            let filename = fileUrl.lastPathComponent
            body.append("--\(boundary)\r\n")
            body.append("Content-Disposition: form-data; name=\"file\"; filename=\"\(filename)\"\r\n")
            body.append("Content-Type: \(mimeType)\r\n\r\n")
            body.append(fileData)
            body.append("\r\n")
        }
        
        body.append("--\(boundary)--\r\n")
        return body
    }
    
    func mimeTypeFor(_ pathExtension: String) -> String? {
        switch pathExtension.lowercased() {
        case "jpg", "jpeg":
            return "image/jpeg"
        case "png":
            return "image/png"
        case "doc":
            return "application/msword"
        case "pdf":
            return "application/pdf"
        case "mp4":
            return "video/mp4"
        default:
            return nil
        }
    }
}


// MARK: - Data extension for appending strings
extension Data {
    mutating func append(_ string: String) {
        if let data = string.data(using: .utf8) {
            append(data)
        }
    }
}


struct ServerErrorResponse: Codable {
    let errors: [ServerErrorDetail]
}

struct ServerErrorDetail: Codable {
    let message: String
    let extensions: [String: String]
}


extension Endpoint.MethodType {
    var methodName: String {
        switch self {
        case .GET:
            return "GET"
        case .POST:
            return "POST"
        case .UPLOAD:
            return "POST"
        case .PATCH:
            return "PATCH"
        case .DELETE:
            return "DELETE"
        }
    }
}
