//
//  RealmManager.swift
//  LNPopupUIExample
//
//  Created by Vamsi Madduluri on 20/12/23.
//

import Foundation
import RealmSwift

class RealmManager {
    static let shared = RealmManager()
    
    private var realm: Realm
    
    private init() {
        let config = Realm.Configuration(
            // Set the new schema version.
            schemaVersion: 0,
            
            // Migration block
            migrationBlock: { migration, oldSchemaVersion in
                // Perform migrations here if oldSchemaVersion < current version
            }
        )
        
        // Set the default Realm configuration
        Realm.Configuration.defaultConfiguration = config
        
        do {
            realm = try Realm(configuration: config)
        } catch {
            fatalError("Realm initialization error: \(error)")
        }
    }
    
    func add<T: Object>(_ object: T, update: Realm.UpdatePolicy = .error, completion: (() -> Void)? = nil, failure: ((Error) -> Void)? = nil) {
        do {
            try realm.write {
                realm.add(object, update: update)
                completion?()
            }
        } catch {
            failure?(error)
        }
    }
    
    
    func add<T: Object>(_ object: T, completion: (() -> Void)? = nil, failure: ((Error) -> Void)? = nil) {
        do {
            try realm.write {
                realm.add(object)
                completion?()
            }
        } catch {
            failure?(error)
        }
    }
    
    func delete<T: Object>(_ object: T, completion: (() -> Void)? = nil, failure: ((Error) -> Void)? = nil) {
        do {
            try realm.write {
                realm.delete(object)
                completion?()
            }
        } catch {
            failure?(error)
        }
    }

    func delete<T: Object>(_ objects: [T], completion: (() -> Void)? = nil, failure: ((Error) -> Void)? = nil) {
        do {
            try realm.write {
                realm.delete(objects)
                completion?()
            }
        } catch {
            failure?(error)
        }
    }

    func deleteAll(completion: (() -> Void)? = nil, failure: ((Error) -> Void)? = nil) {
        do {
            try realm.write {
                realm.deleteAll()
                completion?()
            }
        } catch {
            failure?(error)
        }
    }

    func deleteAll<T: Object>(ofType type: T.Type, completion: (() -> Void)? = nil, failure: ((Error) -> Void)? = nil) {
        do {
            try realm.write {
                realm.delete(realm.objects(type))
                completion?()
            }
        } catch {
            failure?(error)
        }
    }
    
    func getObjects<T: Object>(ofType type: T.Type, filter: NSPredicate? = nil, sortedBy keyPath: String? = nil, ascending: Bool = true) -> Results<T> {
        var results = realm.objects(type)
        if let filter = filter {
            results = results.filter(filter)
        }
        if let keyPath = keyPath {
            results = results.sorted(byKeyPath: keyPath, ascending: ascending)
        }
        return results
    }
    
    func getObject<T: Object, KeyType>(ofType type: T.Type, forPrimaryKey key: KeyType) -> T? {
        return realm.object(ofType: type, forPrimaryKey: key)
    }
    
    func update<T: Object>(_ object: T, with dictionary: [String: Any?], completion: (() -> Void)? = nil, failure: ((Error) -> Void)? = nil) {
        do {
            try realm.write {
                for (key, value) in dictionary {
                    object.setValue(value, forKey: key)
                }
                completion?()
            }
        } catch {
            failure?(error)
        }
    }
}
